#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["create_attribute_from_type"]

from typing import (
    TYPE_CHECKING,
    List,
)

from neptune.api.models import FieldType
from neptune.attributes import (
    Boolean,
    Datetime,
    Float,
    FloatSeries,
    Integer,
    NotebookRef,
    RunState,
    String,
    StringSeries,
    StringSet,
)
from neptune.internal.exceptions import InternalClientError

if TYPE_CHECKING:
    from neptune.attributes.attribute import Attribute
    from neptune.objects import NeptuneObject

_attribute_type_to_attr_class_map = {
    FieldType.FLOAT: Float,
    FieldType.INT: Integer,
    FieldType.BOOL: Boolean,
    FieldType.STRING: String,
    FieldType.DATETIME: Datetime,
    FieldType.FLOAT_SERIES: FloatSeries,
    FieldType.STRING_SERIES: StringSeries,
    FieldType.STRING_SET: StringSet,
    FieldType.OBJECT_STATE: RunState,
    FieldType.NOTEBOOK_REF: NotebookRef,
}


def create_attribute_from_type(
    attribute_type: FieldType,
    container: "NeptuneObject",
    path: List[str],
) -> "Attribute":
    try:
        return _attribute_type_to_attr_class_map[attribute_type](container, path)
    except KeyError:
        raise InternalClientError(f"Unexpected type: {attribute_type}")


def delayed_():
    pass



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["FetchableSeries"]

import abc
from datetime import datetime
from typing import (
    Dict,
    Generic,
    Optional,
    TypeVar,
    Union,
)

from neptune.api.fetching_series_values import fetch_series_values
from neptune.api.models import (
    FloatPointValue,
    StringPointValue,
)
from neptune.internal.utils.paths import path_to_str
from neptune.typing import ProgressBarType

Row = TypeVar("Row", StringPointValue, FloatPointValue)


def make_row(entry: Row, include_timestamp: bool = True) -> Dict[str, Union[str, float, datetime]]:
    row: Dict[str, Union[str, float, datetime]] = {
        "step": entry.step,
        "value": entry.value,
    }

    if include_timestamp:
        row["timestamp"] = entry.timestamp

    return row


class FetchableSeries(Generic[Row]):
    @abc.abstractmethod
    def _fetch_values_from_backend(self, limit: int, from_step: Optional[float] = None) -> Row: ...

    def fetch_values(self, *, include_timestamp: bool = True, progress_bar: Optional[ProgressBarType] = None):
        import pandas as pd

        path = path_to_str(self._path) if hasattr(self, "_path") else ""
        data = fetch_series_values(
            getter=self._fetch_values_from_backend,
            path=path,
            progress_bar=progress_bar,
        )

        rows = dict((n, make_row(entry=entry, include_timestamp=include_timestamp)) for (n, entry) in enumerate(data))

        df = pd.DataFrame.from_dict(data=rows, orient="index")
        return df



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["STOP_QUEUE_STATUS_UPDATE_FREQ_SECONDS", "STOP_QUEUE_MAX_TIME_NO_CONNECTION_SECONDS"]

import os

from neptune.envs import NEPTUNE_SYNC_AFTER_STOP_TIMEOUT
from neptune.internal.parameters import DEFAULT_STOP_TIMEOUT

STOP_QUEUE_STATUS_UPDATE_FREQ_SECONDS = 30.0
STOP_QUEUE_MAX_TIME_NO_CONNECTION_SECONDS = float(os.getenv(NEPTUNE_SYNC_AFTER_STOP_TIMEOUT, DEFAULT_STOP_TIMEOUT))



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class ProtoFloatSeriesValuesDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOTAL_ITEM_COUNT_FIELD_NUMBER: builtins.int
    VALUES_FIELD_NUMBER: builtins.int
    total_item_count: builtins.int
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ProtoFloatPointValueDTO]: ...
    def __init__(
        self,
        *,
        total_item_count: builtins.int = ...,
        values: collections.abc.Iterable[global___ProtoFloatPointValueDTO] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["total_item_count", b"total_item_count", "values", b"values"]) -> None: ...

global___ProtoFloatSeriesValuesDTO = ProtoFloatSeriesValuesDTO

@typing_extensions.final
class ProtoFloatPointValueDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TIMESTAMP_MILLIS_FIELD_NUMBER: builtins.int
    STEP_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    timestamp_millis: builtins.int
    step: builtins.float
    value: builtins.float
    def __init__(
        self,
        *,
        timestamp_millis: builtins.int = ...,
        step: builtins.float = ...,
        value: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["step", b"step", "timestamp_millis", b"timestamp_millis", "value", b"value"]) -> None: ...

global___ProtoFloatPointValueDTO = ProtoFloatPointValueDTO



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import threading
from dataclasses import dataclass
from pathlib import Path
from types import TracebackType
from typing import (
    Any,
    Callable,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

from neptune.core.components.abstract import WithResources
from neptune.core.components.queue.disk_queue import (
    DiskQueue,
    QueueElement,
)

T = TypeVar("T")
K = TypeVar("K")
Timestamp = float


@dataclass
class CategoryQueueElement(Generic[T, K]):
    obj: T
    category: Optional[K] = None


def to_dict_factory(to_dict: Callable[[T], dict]) -> Callable[[CategoryQueueElement[T, K]], dict]:
    def _to_dict(obj: CategoryQueueElement[T, K]) -> dict:
        return {"obj": to_dict(obj.obj), "cat": obj.category}

    return _to_dict


def from_dict_factory(from_dict: Callable[[dict], T]) -> Callable[[dict], CategoryQueueElement[T, Any]]:
    def _from_dict(data: dict) -> CategoryQueueElement[T, Any]:
        return CategoryQueueElement(from_dict(data["obj"]), data["cat"])

    return _from_dict


class AggregatingDiskQueue(WithResources, Generic[T, K]):
    def __init__(
        self,
        data_path: Path,
        to_dict: Callable[[T], dict],
        from_dict: Callable[[dict], T],
        lock: threading.RLock,
        max_file_size: int = 64 * 1024**2,
        max_batch_size_bytes: Optional[int] = None,
        extension: str = "log",
    ) -> None:
        self._disk_queue = DiskQueue[CategoryQueueElement[T, K]](
            data_path=data_path,
            to_dict=to_dict_factory(to_dict),
            from_dict=from_dict_factory(from_dict),
            lock=lock,
            max_file_size=max_file_size,
            max_batch_size_bytes=max_batch_size_bytes,
            extension=extension,
        )
        self._stored_element: Optional[QueueElement[CategoryQueueElement[T, K]]] = None
        self._empty_cond = threading.Condition(threading.Lock())

    def put(self, obj: T, category: Optional[K] = None) -> int:
        return self._disk_queue.put(CategoryQueueElement(obj, category))

    def get(self) -> Optional[QueueElement[CategoryQueueElement[T, K]]]:
        if self._stored_element is not None:
            tmp = self._stored_element
            self._stored_element = None
            return tmp
        else:
            return self._disk_queue.get()

    def get_batch(self, size: int) -> List[QueueElement[CategoryQueueElement[T, K]]]:
        if self._stored_element is not None:
            first: QueueElement[CategoryQueueElement[T, K]] = self._stored_element
            self._stored_element = None
        else:
            possible_first = self._disk_queue.get()
            if not possible_first:
                return []
            else:
                first = possible_first

        ret = [first]
        category = first.obj.category
        cur_batch_size = first.size
        for _ in range(0, size - 1):
            if cur_batch_size >= self._disk_queue._max_batch_size_bytes:
                break
            next_obj = self._disk_queue._get()
            if not next_obj:
                break

            if next_obj.obj.category is not None:
                if category is None:
                    category = next_obj.obj.category
                elif category != next_obj.obj.category:
                    self._stored_element = next_obj
                    break

            cur_batch_size += next_obj.size
            ret.append(next_obj)
        return ret

    def __enter__(self) -> "AggregatingDiskQueue[T, K]":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._disk_queue.__exit__(exc_type, exc_value, traceback)

    def close(self) -> None:
        self._disk_queue.close()

    def flush(self) -> None:
        self._disk_queue.flush()

    def ack(self, version: int) -> None:
        if self._stored_element and self._stored_element.ver <= version:
            self._stored_element = None
        self._disk_queue.ack(version)

    def size(self) -> int:
        return self._disk_queue.size()

    def is_empty(self) -> bool:
        return self._disk_queue.is_empty() and self._stored_element is None

    def cleanup(self) -> None:
        self._disk_queue.cleanup()

    @property
    def resources(self) -> Tuple[DiskQueue[CategoryQueueElement[T, K]]]:
        return (self._disk_queue,)

    @property
    def data_path(self) -> Path:
        return self._disk_queue.data_path

    def wait_for_empty(self, seconds: Optional[float] = None) -> bool:
        return self._disk_queue.wait_for_empty(seconds) and self._empty_cond.wait_for(self.is_empty, timeout=seconds)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["ClearRunner"]

import shutil
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Sequence,
)

import click

from neptune.cli.collect import collect_containers
from neptune.cli.status import StatusRunner
from neptune.constants import SYNC_DIRECTORY
from neptune.internal.utils.logger import get_logger

if TYPE_CHECKING:
    from neptune.cli.containers import (
        AsyncContainer,
        Container,
        OfflineContainer,
    )
    from neptune.internal.backends.neptune_backend import NeptuneBackend

logger = get_logger(with_prefix=False)


class ClearRunner:
    @staticmethod
    def clear(*, backend: "NeptuneBackend", path: Path, force: bool = False, clear_eventual: bool = True) -> None:
        containers = collect_containers(path=path, backend=backend)

        remove_sync_containers(path=path)
        remove_containers(containers=containers.not_found_containers)
        remove_containers(
            containers=filter_containers(a=containers.synced_containers, b=containers.not_found_containers)
        )

        if clear_eventual and (containers.offline_containers or containers.unsynced_containers):
            log_junk_metadata(
                offline_containers=containers.offline_containers, unsynced_containers=containers.unsynced_containers
            )

            if force or click.confirm("\nDo you want to delete the listed metadata?"):
                remove_containers(containers=containers.offline_containers)
                remove_containers(containers=containers.unsynced_containers)


def filter_containers(*, a: Sequence["Container"], b: Sequence["Container"]) -> Sequence["Container"]:
    b_ids = {container.container_id for container in b}
    return [container for container in a if container.container_id not in b_ids]


def remove_sync_containers(*, path: Path) -> None:
    """
    Function can remove SYNC_DIRECTORY safely, Neptune client only stores files to upload in this location.
    """
    shutil.rmtree(path / SYNC_DIRECTORY, ignore_errors=True)


def log_junk_metadata(
    *, offline_containers: Sequence["OfflineContainer"], unsynced_containers: Sequence["AsyncContainer"]
) -> None:
    if unsynced_containers:
        logger.info("")
        StatusRunner.log_unsync_objects(unsynced_containers=unsynced_containers)

    if offline_containers:
        logger.info("")
        StatusRunner.log_offline_objects(offline_containers=offline_containers, info=False)


def remove_containers(*, containers: Sequence["Container"]) -> None:
    for container in containers:
        container.clear()



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "require_installed",
    "is_installed",
]

from functools import lru_cache
from importlib.util import find_spec
from typing import Optional

from neptune.exceptions import NeptuneMissingRequirementException


@lru_cache(maxsize=32)
def is_installed(requirement_name: str) -> bool:
    return find_spec(requirement_name) is not None


def require_installed(requirement_name: str, *, suggestion: Optional[str] = None) -> None:
    if is_installed(requirement_name):
        return

    raise NeptuneMissingRequirementException(requirement_name, suggestion)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["HostedNeptuneBackend"]

import itertools
import os
import re
import typing
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from bravado.exception import (
    HTTPConflict,
    HTTPNotFound,
    HTTPPaymentRequired,
    HTTPUnprocessableEntity,
)

from neptune.api.models import (
    BoolField,
    DateTimeField,
    Field,
    FieldDefinition,
    FieldType,
    FloatField,
    FloatSeriesField,
    FloatSeriesValues,
    IntField,
    LeaderboardEntry,
    NextPage,
    QueryFieldDefinitionsResult,
    QueryFieldsResult,
    StringField,
    StringSeriesField,
    StringSeriesValues,
    StringSetField,
)
from neptune.api.proto.neptune_pb.api.model.attributes_pb2 import ProtoAttributesSearchResultDTO
from neptune.api.proto.neptune_pb.api.model.leaderboard_entries_pb2 import ProtoAttributesDTO
from neptune.api.proto.neptune_pb.api.model.series_values_pb2 import ProtoFloatSeriesValuesDTO
from neptune.api.searching_entries import iter_over_pages
from neptune.core.components.operation_storage import OperationStorage
from neptune.envs import (
    NEPTUNE_FETCH_TABLE_STEP_SIZE,
    NEPTUNE_USE_PROTOCOL_BUFFERS,
)
from neptune.exceptions import (
    AmbiguousProjectName,
    ContainerUUIDNotFound,
    FetchAttributeNotFoundException,
    MetadataContainerNotFound,
    MetadataInconsistency,
    NeptuneFeatureNotAvailableException,
    NeptuneLimitExceedException,
    NeptuneObjectCreationConflict,
    ProjectNotFound,
    ProjectNotFoundWithSuggestions,
)
from neptune.internal.backends.api_model import (
    ApiExperiment,
    Project,
    Workspace,
)
from neptune.internal.backends.hosted_client import (
    DEFAULT_PROTO_REQUEST_KWARGS,
    DEFAULT_REQUEST_KWARGS,
    create_backend_client,
    create_http_client_with_auth,
    create_leaderboard_client,
)
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.backends.nql import NQLQuery
from neptune.internal.backends.operation_api_name_visitor import OperationApiNameVisitor
from neptune.internal.backends.operation_api_object_converter import OperationApiObjectConverter
from neptune.internal.backends.operations_preprocessor import OperationsPreprocessor
from neptune.internal.backends.utils import (
    ExecuteOperationsBatchingManager,
    build_operation_url,
    ssl_verify,
    with_api_exceptions_handler,
)
from neptune.internal.container_type import ContainerType
from neptune.internal.credentials import Credentials
from neptune.internal.exceptions import NeptuneException
from neptune.internal.id_formats import (
    QualifiedName,
    UniqueId,
)
from neptune.internal.operation import Operation
from neptune.internal.utils.generic_attribute_mapper import map_attribute_result_to_value
from neptune.internal.utils.logger import get_logger
from neptune.internal.utils.paths import path_to_str
from neptune.internal.utils.patterns import PROJECT_QUALIFIED_NAME_PATTERN
from neptune.internal.warnings import (
    NeptuneWarning,
    warn_once,
)
from neptune.internal.websockets.websockets_factory import WebsocketsFactory
from neptune.typing import ProgressBarType
from neptune.version import __version__

if TYPE_CHECKING:
    from bravado.requests_client import RequestsClient

    from neptune.internal.backends.api_model import ClientConfig


_logger = get_logger()

ATOMIC_ATTRIBUTE_TYPES = {
    FieldType.INT.value,
    FieldType.FLOAT.value,
    FieldType.STRING.value,
    FieldType.BOOL.value,
    FieldType.DATETIME.value,
    FieldType.OBJECT_STATE.value,
}


class HostedNeptuneBackend(NeptuneBackend):
    def __init__(self, credentials: Credentials, proxies: Optional[Dict[str, str]] = None):
        self.credentials = credentials
        self.proxies = proxies
        self.missing_features = []
        self.use_proto = os.getenv(NEPTUNE_USE_PROTOCOL_BUFFERS, "False").lower() in {"true", "1", "y"}

        http_client, client_config = create_http_client_with_auth(
            credentials=credentials, ssl_verify=ssl_verify(), proxies=proxies
        )
        self._http_client: "RequestsClient" = http_client
        self._client_config: "ClientConfig" = client_config

        self.backend_client = create_backend_client(self._client_config, self._http_client)
        self.leaderboard_client = create_leaderboard_client(self._client_config, self._http_client)

    def verify_feature_available(self, feature_name: str):
        if not self._client_config.has_feature(feature_name):
            raise NeptuneFeatureNotAvailableException(feature_name)

    def websockets_factory(self, project_id: str, run_id: str) -> Optional[WebsocketsFactory]:
        base_url = re.sub(r"^http", "ws", self._client_config.api_url)
        return WebsocketsFactory(
            url=build_operation_url(base_url, f"/api/notifications/v1/runs/{project_id}/{run_id}/signal"),
            session=self._http_client.authenticator.auth.session,
            proxies=self.proxies,
        )

    @with_api_exceptions_handler
    def get_project(self, project_id: QualifiedName) -> Project:
        project_spec = re.search(PROJECT_QUALIFIED_NAME_PATTERN, project_id)
        workspace, name = project_spec["workspace"], project_spec["project"]

        try:
            if not workspace:
                available_projects = list(
                    filter(
                        lambda p: p.name == name,
                        self.get_available_projects(search_term=name),
                    )
                )

                if len(available_projects) == 1:
                    project = available_projects[0]
                    project_id = f"{project.workspace}/{project.name}"
                elif len(available_projects) > 1:
                    raise AmbiguousProjectName(project_id=project_id, available_projects=available_projects)
                else:
                    raise ProjectNotFoundWithSuggestions(
                        project_id=project_id,
                        available_projects=self.get_available_projects(),
                        available_workspaces=self.get_available_workspaces(),
                    )

            response = self.backend_client.api.getProject(
                projectIdentifier=project_id,
                **DEFAULT_REQUEST_KWARGS,
            ).response()
            project = response.result

            return Project(
                id=project.id,
                name=project.name,
                workspace=project.organizationName,
                sys_id=project.projectKey,
            )
        except HTTPNotFound:
            available_workspaces = self.get_available_workspaces()

            if workspace and not list(filter(lambda aw: aw.name == workspace, available_workspaces)):
                # Could not found specified workspace, forces listing all projects
                raise ProjectNotFoundWithSuggestions(
                    project_id=project_id,
                    available_projects=self.get_available_projects(),
                    available_workspaces=available_workspaces,
                )
            else:
                raise ProjectNotFoundWithSuggestions(
                    project_id=project_id,
                    available_projects=self.get_available_projects(workspace_id=workspace),
                )

    @with_api_exceptions_handler
    def get_available_projects(
        self, workspace_id: Optional[str] = None, search_term: Optional[str] = None
    ) -> List[Project]:
        try:
            response = self.backend_client.api.listProjects(
                limit=5,
                organizationIdentifier=workspace_id,
                searchTerm=search_term,
                sortBy=["lastViewed"],
                sortDirection=["descending"],
                userRelation="memberOrHigher",
                **DEFAULT_REQUEST_KWARGS,
            ).response()
            projects = response.result.entries
            return list(
                map(
                    lambda project: Project(
                        id=project.id,
                        name=project.name,
                        workspace=project.organizationName,
                        sys_id=project.projectKey,
                    ),
                    projects,
                )
            )
        except HTTPNotFound:
            return []

    @with_api_exceptions_handler
    def get_available_workspaces(self) -> List[Workspace]:
        try:
            response = self.backend_client.api.listOrganizations(
                **DEFAULT_REQUEST_KWARGS,
            ).response()
            workspaces = response.result
            return list(
                map(
                    lambda workspace: Workspace(id=workspace.id, name=workspace.name),
                    workspaces,
                )
            )
        except HTTPNotFound:
            return []

    @with_api_exceptions_handler
    def get_metadata_container(
        self,
        container_id: Union[UniqueId, QualifiedName],
        expected_container_type: typing.Optional[ContainerType],
    ) -> ApiExperiment:
        experiment = (
            self.leaderboard_client.api.getExperiment(
                experimentId=container_id,
                **DEFAULT_REQUEST_KWARGS,
            )
            .response()
            .result
        )

        if expected_container_type is not None and ContainerType.from_api(experiment.type) != expected_container_type:
            raise MetadataContainerNotFound.of_container_type(
                container_type=expected_container_type, container_id=container_id
            )

        return ApiExperiment.from_experiment(experiment)

    @with_api_exceptions_handler
    def create_run(
        self,
        project_id: UniqueId,
        custom_run_id: Optional[str] = None,
        notebook_id: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
    ) -> ApiExperiment:
        additional_params = {
            "customId": custom_run_id,
        }

        if notebook_id is not None and checkpoint_id is not None:
            additional_params["notebookId"] = notebook_id if notebook_id is not None else None
            additional_params["checkpointId"] = checkpoint_id if checkpoint_id is not None else None

        return self._create_experiment(
            project_id=project_id,
            parent_id=project_id,
            container_type=ContainerType.RUN,
            additional_params=additional_params,
        )

    @with_api_exceptions_handler
    def create_model(self, project_id: UniqueId, key: str = "") -> ApiExperiment:
        additional_params = {
            "key": key,
        }

        return self._create_experiment(
            project_id=project_id,
            parent_id=project_id,
            container_type=ContainerType.MODEL,
            additional_params=additional_params,
        )

    @with_api_exceptions_handler
    def create_model_version(self, project_id: UniqueId, model_id: UniqueId) -> ApiExperiment:
        return self._create_experiment(
            project_id=project_id,
            parent_id=model_id,
            container_type=ContainerType.MODEL_VERSION,
        )

    def _create_experiment(
        self,
        project_id: UniqueId,
        parent_id: UniqueId,
        container_type: ContainerType,
        additional_params: Optional[dict] = None,
    ):
        if additional_params is None:
            additional_params = dict()

        params = {
            "projectIdentifier": project_id,
            "parentId": parent_id,
            "type": container_type.to_api(),
            "cliVersion": __version__,
            **additional_params,
        }

        kwargs = {
            "experimentCreationParams": params,
            "X-Neptune-CliVersion": __version__,
            **DEFAULT_REQUEST_KWARGS,
        }

        try:
            experiment = self.leaderboard_client.api.createExperiment(**kwargs).response().result
            return ApiExperiment.from_experiment(experiment)
        except HTTPNotFound:
            raise ProjectNotFound(project_id=project_id)
        except HTTPConflict as e:
            raise NeptuneObjectCreationConflict() from e

    @with_api_exceptions_handler
    def ping(self, container_id: str, container_type: ContainerType):
        request_kwargs = {
            "_request_options": {
                "timeout": 10,
                "connect_timeout": 10,
            }
        }
        try:
            self.leaderboard_client.api.ping(
                experimentId=container_id,
                **request_kwargs,
            ).response().result
        except HTTPNotFound as e:
            raise ContainerUUIDNotFound(container_id, container_type) from e

    def execute_operations(
        self,
        container_id: UniqueId,
        container_type: ContainerType,
        operations: List[Operation],
        operation_storage: OperationStorage,
    ) -> Tuple[int, List[NeptuneException]]:
        errors = []

        batching_mgr = ExecuteOperationsBatchingManager(self)
        operations_batch = batching_mgr.get_batch(operations)
        errors.extend(operations_batch.errors)
        dropped_count = operations_batch.dropped_operations_count

        operations_preprocessor = OperationsPreprocessor()
        operations_preprocessor.process(operations_batch.operations)

        preprocessed_operations = operations_preprocessor.get_operations()
        errors.extend(preprocessed_operations.errors)

        errors.extend(
            self._execute_operations(
                container_id,
                container_type,
                operations=preprocessed_operations.other_operations,
            )
        )

        for op in itertools.chain(
            preprocessed_operations.upload_operations,
            preprocessed_operations.other_operations,
        ):
            op.clean(operation_storage=operation_storage)

        return (
            operations_preprocessor.processed_ops_count + dropped_count,
            errors,
        )

    @with_api_exceptions_handler
    def _execute_operations(
        self,
        container_id: UniqueId,
        container_type: ContainerType,
        operations: List[Operation],
    ) -> List[MetadataInconsistency]:
        kwargs = {
            "experimentId": container_id,
            "operations": [
                {
                    "path": path_to_str(op.path),
                    OperationApiNameVisitor().visit(op): OperationApiObjectConverter().convert(op),
                }
                for op in operations
            ],
            **DEFAULT_REQUEST_KWARGS,
        }

        try:
            result = self.leaderboard_client.api.executeOperations(**kwargs).response().result
            return [MetadataInconsistency(err.errorDescription) for err in result]
        except HTTPNotFound as e:
            raise ContainerUUIDNotFound(container_id, container_type) from e
        except (HTTPPaymentRequired, HTTPUnprocessableEntity) as e:
            raise NeptuneLimitExceedException(reason=e.response.json().get("title", "Unknown reason")) from e

    @with_api_exceptions_handler
    def get_attributes(self, container_id: str, container_type: ContainerType) -> List[FieldDefinition]:
        params = {
            "experimentId": container_id,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            experiment = self.leaderboard_client.api.getExperimentAttributes(**params).response().result

            attribute_type_names = [at.value for at in FieldType]
            accepted_attributes = [attr for attr in experiment.attributes if attr.type in attribute_type_names]

            # Notify about ignored attrs
            ignored_attributes = set(attr.type for attr in experiment.attributes) - set(
                attr.type for attr in accepted_attributes
            )
            if ignored_attributes:
                _logger.warning(
                    "Ignored following attributes (unknown type): %s.\n" "Try to upgrade `neptune`.",
                    ignored_attributes,
                )

            return [
                FieldDefinition.from_model(field) for field in accepted_attributes if field.type in attribute_type_names
            ]
        except HTTPNotFound as e:
            raise ContainerUUIDNotFound(
                container_id=container_id,
                container_type=container_type,
            ) from e

    @with_api_exceptions_handler
    def get_float_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> FloatField:
        params = {
            "experimentId": container_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getFloatAttribute(**params).response().result
            return FloatField.from_model(result)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_int_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> IntField:
        params = {
            "experimentId": container_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getIntAttribute(**params).response().result
            return IntField.from_model(result)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_bool_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> BoolField:
        params = {
            "experimentId": container_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getBoolAttribute(**params).response().result
            return BoolField.from_model(result)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_string_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> StringField:
        params = {
            "experimentId": container_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getStringAttribute(**params).response().result
            return StringField.from_model(result)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_datetime_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> DateTimeField:
        params = {
            "experimentId": container_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getDatetimeAttribute(**params).response().result
            return DateTimeField.from_model(result)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_float_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> FloatSeriesField:
        params = {
            "experimentId": container_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getFloatSeriesAttribute(**params).response().result
            return FloatSeriesField.from_model(result)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_string_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSeriesField:
        params = {
            "experimentId": container_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getStringSeriesAttribute(**params).response().result
            return StringSeriesField.from_model(result)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_string_set_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSetField:
        params = {
            "experimentId": container_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getStringSetAttribute(**params).response().result
            return StringSetField.from_model(result)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_string_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        limit: int,
        from_step: Optional[float] = None,
    ) -> StringSeriesValues:
        params = {
            "experimentId": container_id,
            "attribute": path_to_str(path),
            "limit": limit,
            "skipToStep": from_step,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = self.leaderboard_client.api.getStringSeriesValues(**params).response().result
            return StringSeriesValues.from_model(result)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_float_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        limit: int,
        from_step: Optional[float] = None,
        use_proto: Optional[bool] = None,
    ) -> FloatSeriesValues:
        use_proto = use_proto if use_proto is not None else self.use_proto

        params = {
            "experimentId": container_id,
            "attribute": path_to_str(path),
            "limit": limit,
            "skipToStep": from_step,
        }
        try:
            if use_proto:
                result = (
                    self.leaderboard_client.api.getFloatSeriesValuesProto(
                        **params,
                        **DEFAULT_PROTO_REQUEST_KWARGS,
                    )
                    .response()
                    .result
                )
                data = ProtoFloatSeriesValuesDTO.FromString(result)
                return FloatSeriesValues.from_proto(data)
            else:
                result = (
                    self.leaderboard_client.api.getFloatSeriesValues(
                        **params,
                        **DEFAULT_REQUEST_KWARGS,
                    )
                    .response()
                    .result
                )
                return FloatSeriesValues.from_model(result)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def query_fields_within_project(
        self,
        project_id: QualifiedName,
        field_names_filter: Optional[List[str]] = None,
        experiment_ids_filter: Optional[List[str]] = None,
        next_page: Optional[NextPage] = None,
    ) -> QueryFieldsResult:
        pagination = {"nextPage": next_page.to_dto()} if next_page else {}
        params = {
            "projectIdentifier": project_id,
            "query": {
                **pagination,
                "attributeNamesFilter": field_names_filter,
                "experimentIdsFilter": experiment_ids_filter,
            },
            **DEFAULT_REQUEST_KWARGS,
        }

        try:
            result = self.leaderboard_client.api.queryAttributesWithinProject(**params).response().result
            return QueryFieldsResult.from_model(result)
        except HTTPNotFound:
            raise ProjectNotFound(project_id=project_id)

    @with_api_exceptions_handler
    def fetch_atom_attribute_values(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> List[Tuple[str, FieldType, Any]]:
        params = {
            "experimentId": container_id,
        }
        try:
            namespace_prefix = path_to_str(path)
            if namespace_prefix:
                # don't want to catch "ns/attribute/other" while looking for "ns/attr"
                namespace_prefix += "/"
            result = self.leaderboard_client.api.getExperimentAttributes(**params).response().result
            return [
                (attr.name, attr.type, map_attribute_result_to_value(attr))
                for attr in result.attributes
                if attr.name.startswith(namespace_prefix)
            ]
        except HTTPNotFound as e:
            raise ContainerUUIDNotFound(container_id, container_type) from e

    @with_api_exceptions_handler
    def _get_column_types(self, project_id: UniqueId, column: str, types: Optional[Iterable[str]] = None) -> List[Any]:
        params = {
            "projectIdentifier": project_id,
            "search": column,
            "type": types,
            "params": {},
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            return self.leaderboard_client.api.searchLeaderboardAttributes(**params).response().result.entries
        except HTTPNotFound as e:
            raise ProjectNotFound(project_id=project_id) from e

    @with_api_exceptions_handler
    def search_leaderboard_entries(
        self,
        project_id: UniqueId,
        types: Optional[Iterable[ContainerType]] = None,
        query: Optional[NQLQuery] = None,
        columns: Optional[Iterable[str]] = None,
        limit: Optional[int] = None,
        sort_by: str = "sys/creation_time",
        ascending: bool = False,
        progress_bar: Optional[ProgressBarType] = None,
        step_size: Optional[int] = None,
        use_proto: Optional[bool] = None,
    ) -> Generator[LeaderboardEntry, None, None]:
        use_proto = use_proto if use_proto is not None else self.use_proto
        default_step_size = step_size or int(os.getenv(NEPTUNE_FETCH_TABLE_STEP_SIZE, "100"))

        step_size = min(default_step_size, limit) if limit else default_step_size

        types_filter = list(map(lambda container_type: container_type.to_api(), types)) if types else None
        attributes_filter = {"attributeFilters": [{"path": column} for column in columns]} if columns else {}

        if sort_by == "sys/creation_time":
            sort_by_column_type = FieldType.DATETIME.value
        elif sort_by == "sys/id":
            sort_by_column_type = FieldType.STRING.value
        else:
            sort_by_column_type_candidates = self._get_column_types(project_id, sort_by, types_filter)
            sort_by_column_type = _get_column_type_from_entries(sort_by_column_type_candidates, sort_by)

        try:
            return iter_over_pages(
                client=self.leaderboard_client,
                project_id=project_id,
                types=types_filter,
                query=query,
                attributes_filter=attributes_filter,
                step_size=step_size,
                limit=limit,
                sort_by=sort_by,
                ascending=ascending,
                sort_by_column_type=sort_by_column_type,
                progress_bar=progress_bar,
                use_proto=use_proto,
            )
        except HTTPNotFound:
            raise ProjectNotFound(project_id)

    def query_fields_definitions_within_project(
        self,
        project_id: QualifiedName,
        field_name_regex: Optional[str] = None,
        experiment_ids_filter: Optional[List[str]] = None,
        next_page: Optional[NextPage] = None,
    ) -> QueryFieldDefinitionsResult:
        pagination = {"nextPage": next_page.to_dto()} if next_page else {}
        params = {
            "projectIdentifier": project_id,
            "query": {
                **pagination,
                "experimentIdsFilter": experiment_ids_filter,
                "attributeNameRegex": field_name_regex,
            },
        }

        try:
            data = (
                self.leaderboard_client.api.queryAttributeDefinitionsWithinProject(
                    **params,
                    **DEFAULT_REQUEST_KWARGS,
                )
                .response()
                .result
            )
            return QueryFieldDefinitionsResult.from_model(data)
        except HTTPNotFound:
            raise ProjectNotFound(project_id=project_id)

    def get_fields_definitions(
        self,
        container_id: str,
        container_type: ContainerType,
        use_proto: Optional[bool] = None,
    ) -> List[FieldDefinition]:
        use_proto = use_proto if use_proto is not None else self.use_proto

        params = {
            "experimentIdentifier": container_id,
        }

        try:
            if use_proto:
                result = (
                    self.leaderboard_client.api.queryAttributeDefinitionsProto(
                        **params,
                        **DEFAULT_PROTO_REQUEST_KWARGS,
                    )
                    .response()
                    .result
                )
                data = ProtoAttributesSearchResultDTO.FromString(result)
                return [FieldDefinition.from_proto(field_def) for field_def in data.entries]
            else:
                data = (
                    self.leaderboard_client.api.queryAttributeDefinitions(
                        **params,
                        **DEFAULT_REQUEST_KWARGS,
                    )
                    .response()
                    .result
                )
                return [FieldDefinition.from_model(field_def) for field_def in data.entries]
        except HTTPNotFound as e:
            raise ContainerUUIDNotFound(
                container_id=container_id,
                container_type=container_type,
            ) from e

    def get_fields_with_paths_filter(
        self, container_id: str, container_type: ContainerType, paths: List[str], use_proto: Optional[bool] = None
    ) -> List[Field]:
        use_proto = use_proto if use_proto is not None else self.use_proto

        params = {
            "holderIdentifier": container_id,
            "holderType": "experiment",
            "attributeQuery": {
                "attributePathsFilter": paths,
            },
        }

        try:
            if use_proto:
                result = (
                    self.leaderboard_client.api.getAttributesWithPathsFilterProto(
                        **params,
                        **DEFAULT_PROTO_REQUEST_KWARGS,
                    )
                    .response()
                    .result
                )
                data = ProtoAttributesDTO.FromString(result)
                return [Field.from_proto(field) for field in data.attributes]
            else:
                data = (
                    self.leaderboard_client.api.getAttributesWithPathsFilter(
                        **params,
                        **DEFAULT_REQUEST_KWARGS,
                    )
                    .response()
                    .result
                )
                return [Field.from_model(field) for field in data.attributes]
        except HTTPNotFound as e:
            raise ContainerUUIDNotFound(
                container_id=container_id,
                container_type=container_type,
            ) from e


def _get_column_type_from_entries(entries: List[Any], column: str) -> str:
    if not entries:  # column chosen is not present in the table
        raise ValueError(f"Column '{column}' chosen for sorting is not present in the table")

    if len(entries) == 1 and entries[0].name == column:
        return entries[0].type

    types = set()
    for entry in entries:
        if entry.name != column:  # caught by regex, but it's not this column
            continue
        if entry.type not in ATOMIC_ATTRIBUTE_TYPES:  # non-atomic type - no need to look further
            raise ValueError(
                f"Column {column} used for sorting is a complex type. For more, "
                f"see https://docs.neptune.ai/api/field_types/#simple-types"
            )
        types.add(entry.type)

    if types == {FieldType.INT.value, FieldType.FLOAT.value}:
        return FieldType.FLOAT.value

    warn_once(
        f"Column {column} contains more than one simple data type. Sorting result might be inaccurate.",
        exception=NeptuneWarning,
    )
    return FieldType.STRING.value



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["ConsumerThread"]

from time import time
from typing import (
    List,
    Optional,
)

from neptune.core.operation_processors.async_operation_processor.processing_resources import ProcessingResources
from neptune.core.operations.operation import Operation
from neptune.internal.daemon import Daemon
from neptune.internal.signals_processing.utils import (
    signal_batch_lag,
    signal_batch_processed,
    signal_batch_started,
)


class ConsumerThread(Daemon):
    def __init__(
        self,
        sleep_time: float,
        processing_resources: ProcessingResources,
    ) -> None:
        super().__init__(sleep_time=sleep_time, name="NeptuneAsyncOpProcessor")
        self._processing_resources = processing_resources
        self._last_flush: float = 0.0

    def run(self) -> None:
        try:
            super().run()
        except Exception as e:
            with self._processing_resources.waiting_cond:
                self._processing_resources.waiting_cond.notify_all()
            raise Exception from e

    def work(self) -> None:
        ts = time()
        if ts - self._last_flush >= self._sleep_time:
            self._last_flush = ts
            self._processing_resources.disk_queue.flush()

        while True:
            batch = self._processing_resources.disk_queue.get_batch(self._processing_resources.batch_size)
            if not batch:
                return

            signal_batch_started(queue=self._processing_resources.signals_queue)
            self.process_batch([element.obj for element in batch], batch[-1].ver, batch[-1].at)

    @Daemon.ConnectionRetryWrapper(
        kill_message=(
            "Killing Neptune asynchronous thread. All data is safe on disk and can be later"
            " synced manually using `neptune sync` command."
        )
    )
    def process_batch(self, batch: List[Operation], version: int, occurred_at: Optional[float] = None) -> None:
        if occurred_at is not None:
            signal_batch_lag(queue=self._processing_resources.signals_queue, lag=time() - occurred_at)

        expected_count = len(batch)
        version_to_ack = version - expected_count
        while True:
            # TODO: Add proper batch processing logic here once backend is ready

            signal_batch_processed(queue=self._processing_resources.signals_queue)
            processed_count = len(batch)
            version_to_ack += processed_count

            with self._processing_resources.waiting_cond:
                self._processing_resources.disk_queue.ack(version_to_ack)

                self._processing_resources.consumed_version = version_to_ack

                if version_to_ack == version:
                    self._processing_resources.waiting_cond.notify_all()
                    return



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["SignalsProcessor"]

from queue import (
    Empty,
    Queue,
)
from threading import Thread
from time import monotonic
from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
)

from neptune.internal.daemon import Daemon
from neptune.internal.parameters import IN_BETWEEN_CALLBACKS_MINIMUM_INTERVAL
from neptune.internal.signals_processing.signals import (
    BatchLagSignal,
    SignalsVisitor,
)

if TYPE_CHECKING:
    from neptune.internal.signals_processing.signals import Signal
    from neptune.objects import NeptuneObject


class SignalsProcessor(Daemon, SignalsVisitor):
    def __init__(
        self,
        *,
        period: float,
        container: "NeptuneObject",
        queue: "Queue[Signal]",
        async_lag_threshold: float,
        async_no_progress_threshold: float,
        async_lag_callback: Optional[Callable[["NeptuneObject"], None]] = None,
        async_no_progress_callback: Optional[Callable[["NeptuneObject"], None]] = None,
        callbacks_interval: float = IN_BETWEEN_CALLBACKS_MINIMUM_INTERVAL,
        in_async: bool = True,
    ) -> None:
        super().__init__(sleep_time=period, name="CallbacksMonitor")

        self._container: "NeptuneObject" = container
        self._queue: "Queue[Signal]" = queue
        self._async_lag_threshold: float = async_lag_threshold
        self._async_no_progress_threshold: float = async_no_progress_threshold
        self._async_lag_callback: Optional[Callable[["NeptuneObject"], None]] = async_lag_callback
        self._async_no_progress_callback: Optional[Callable[["NeptuneObject"], None]] = async_no_progress_callback
        self._callbacks_interval: float = callbacks_interval
        self._in_async: bool = in_async

        self._last_batch_started_at: Optional[float] = None
        self._last_no_progress_callback_at: Optional[float] = None
        self._last_lag_callback_at: Optional[float] = None

    def visit_batch_started(self, signal: "Signal") -> None:
        if self._last_batch_started_at is None:
            self._last_batch_started_at = signal.occured_at

    def visit_batch_processed(self, signal: "Signal") -> None:
        if self._last_batch_started_at is not None:
            self._check_no_progress(at_timestamp=signal.occured_at)
            self._last_batch_started_at = None

    def visit_batch_lag(self, signal: "Signal") -> None:
        if self._async_lag_callback is None or not isinstance(signal, BatchLagSignal):
            return

        if signal.lag > self._async_lag_threshold:
            current_time = monotonic()
            if (
                self._last_lag_callback_at is None
                or current_time - self._last_lag_callback_at > self._callbacks_interval
            ):
                execute_callback(callback=self._async_lag_callback, container=self._container, in_async=self._in_async)
                self._last_lag_callback_at = current_time

    def _check_callbacks(self) -> None:
        self._check_no_progress(at_timestamp=monotonic())

    def _check_no_progress(self, at_timestamp: float) -> None:
        if self._async_no_progress_callback is None:
            return

        if self._last_batch_started_at is not None:
            if at_timestamp - self._last_batch_started_at > self._async_no_progress_threshold:
                if (
                    self._last_no_progress_callback_at is None
                    or at_timestamp - self._last_no_progress_callback_at > self._callbacks_interval
                ):
                    execute_callback(
                        callback=self._async_no_progress_callback, container=self._container, in_async=self._in_async
                    )
                    self._last_no_progress_callback_at = monotonic()

    def work(self) -> None:
        try:
            while not self._queue.empty():
                signal = self._queue.get_nowait()
                signal.accept(self)
            self._check_callbacks()
        except Empty:
            pass


def execute_callback(
    *, callback: Callable[["NeptuneObject"], None], container: "NeptuneObject", in_async: bool
) -> None:
    if in_async:
        Thread(target=callback, name="CallbackExecution", args=(container,), daemon=True).start()
    else:
        callback(container)



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["get_operation_processor"]

import os
import threading
from queue import Queue
from typing import TYPE_CHECKING

from neptune.core.operation_processors.async_operation_processor import AsyncOperationProcessor
from neptune.core.operation_processors.offline_operation_processor import OfflineOperationProcessor
from neptune.core.operation_processors.operation_processor import OperationProcessor
from neptune.core.operation_processors.read_only_operation_processor import ReadOnlyOperationProcessor
from neptune.core.operation_processors.sync_operation_processor import SyncOperationProcessor
from neptune.core.typing.container_type import ContainerType
from neptune.core.typing.id_formats import CustomId
from neptune.envs import NEPTUNE_ASYNC_BATCH_SIZE
from neptune.objects.mode import Mode

if TYPE_CHECKING:
    from neptune.internal.signals_processing.signals import Signal


def get_operation_processor(
    mode: Mode,
    custom_id: CustomId,
    container_type: ContainerType,
    lock: threading.RLock,
    flush_period: float,
    queue: "Queue[Signal]",
) -> OperationProcessor:
    if mode == Mode.ASYNC:
        return AsyncOperationProcessor(
            custom_id=custom_id,
            container_type=container_type,
            lock=lock,
            sleep_time=flush_period,
            batch_size=int(os.environ.get(NEPTUNE_ASYNC_BATCH_SIZE) or "1000"),
            signal_queue=queue,
        )
    elif mode in {Mode.SYNC, Mode.DEBUG}:
        return SyncOperationProcessor(custom_id=custom_id, container_type=container_type)
    elif mode == Mode.OFFLINE:
        return OfflineOperationProcessor(custom_id=custom_id, container_type=container_type, lock=lock)
    elif mode == Mode.READ_ONLY:
        return ReadOnlyOperationProcessor()
    else:
        raise ValueError(f"mode should be one of {[m for m in Mode]}")



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["WithBackend"]

import abc
from contextlib import AbstractContextManager
from types import TracebackType
from typing import (
    Optional,
    Type,
)

from neptune.internal.backends.api_model import Project
from neptune.internal.backends.factory import get_backend
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.backends.project_name_lookup import project_name_lookup
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import (
    QualifiedName,
    UniqueId,
    conform_optional,
)
from neptune.internal.utils import verify_type
from neptune.objects.mode import Mode


class WithBackend(AbstractContextManager, abc.ABC):
    container_type: ContainerType

    def __init__(
        self,
        api_token: Optional[str] = None,
        project: Optional[str] = None,
        mode: Mode = Mode.ASYNC,
        proxies: Optional[dict] = None,
    ) -> None:
        verify_type("api_token", api_token, (str, type(None)))
        verify_type("mode", mode, Mode)
        verify_type("proxies", proxies, (dict, type(None)))
        verify_type("project", project, (str, type(None)))

        self._mode = mode
        self._backend: NeptuneBackend = get_backend(mode=mode, api_token=api_token, proxies=proxies)
        self._project_qualified_name: Optional[str] = conform_optional(project, QualifiedName)
        self._project_api_object: Project = project_name_lookup(
            backend=self._backend,
            name=QualifiedName(self._project_qualified_name) if self._project_qualified_name is not None else None,
        )
        self._workspace: str = self._project_api_object.workspace
        self._project_name: str = self._project_api_object.name
        self._project_id: UniqueId = self._project_api_object.id

    def close(self) -> None:
        self._backend.close()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "Serializable",
    "Run",
    "FloatValue",
    "LogFloats",
    "AssignInteger",
    "AssignFloat",
    "AssignBool",
    "AssignString",
    "AssignDatetime",
    "RunOperation",
]


import abc
from dataclasses import dataclass
from datetime import datetime
from typing import (
    List,
    Optional,
)

import neptune_api.proto.neptune_pb.ingest.v1.common_pb2 as common_pb2
import neptune_api.proto.neptune_pb.ingest.v1.pub.ingest_pb2 as ingest_pb2
from google.protobuf import timestamp_pb2


class Serializable(abc.ABC):
    @abc.abstractmethod
    def to_proto(self, run_op: "RunOperation") -> ingest_pb2.RunOperation:
        pass


@dataclass
class Run(Serializable):
    created_at: datetime
    custom_id: str

    def to_proto(self, run_op: "RunOperation") -> ingest_pb2.RunOperation:
        return ingest_pb2.RunOperation(
            project=run_op.project,
            run_id=run_op.run_id,
            create=common_pb2.Run(
                creation_time=timestamp_pb2.Timestamp(seconds=int(self.created_at.timestamp())),
                run_id=self.custom_id,
                experiment_id=self.custom_id,
                family=self.custom_id,
            ),
        )


@dataclass
class FloatValue:
    timestamp: float
    value: float
    step: Optional[float] = None


@dataclass
class LogFloats(Serializable):
    path: str
    items: List[FloatValue]

    def to_proto(self, run_op: "RunOperation") -> ingest_pb2.RunOperation:
        first_item = self.items[0]

        step = (
            common_pb2.Step(whole=int(first_item.step), micro=int((first_item.step - int(first_item.step)) * 1e6))
            if first_item.step is not None
            else None
        )

        return ingest_pb2.RunOperation(
            project=run_op.project,
            run_id=run_op.run_id,
            update=common_pb2.UpdateRunSnapshot(
                step=step,
                append={f"{self.path}": common_pb2.Value(float64=first_item.value)},
                timestamp=timestamp_pb2.Timestamp(seconds=int(first_item.timestamp)),
            ),
        )


@dataclass
class AssignInteger(Serializable):
    path: str
    value: int

    def to_proto(self, run_op: "RunOperation") -> ingest_pb2.RunOperation:
        return ingest_pb2.RunOperation(
            project=run_op.project,
            run_id=run_op.run_id,
            update=common_pb2.UpdateRunSnapshot(
                assign={f"{self.path}": common_pb2.Value(int64=self.value)},
            ),
        )


@dataclass
class AssignFloat(Serializable):
    path: str
    value: float

    def to_proto(self, run_op: "RunOperation") -> ingest_pb2.RunOperation:
        return ingest_pb2.RunOperation(
            project=run_op.project,
            run_id=run_op.run_id,
            update=common_pb2.UpdateRunSnapshot(
                assign={f"{self.path}": common_pb2.Value(float64=self.value)},
            ),
        )


@dataclass
class AssignBool(Serializable):
    path: str
    value: bool

    def to_proto(self, run_op: "RunOperation") -> ingest_pb2.RunOperation:
        return ingest_pb2.RunOperation(
            project=run_op.project,
            run_id=run_op.run_id,
            update=common_pb2.UpdateRunSnapshot(
                assign={f"{self.path}": common_pb2.Value(bool=self.value)},
            ),
        )


@dataclass
class AssignString(Serializable):
    path: str
    value: str

    def to_proto(self, run_op: "RunOperation") -> ingest_pb2.RunOperation:
        return ingest_pb2.RunOperation(
            project=run_op.project,
            run_id=run_op.run_id,
            update=common_pb2.UpdateRunSnapshot(
                assign={f"{self.path}": common_pb2.Value(string=self.value)},
            ),
        )


@dataclass
class AssignDatetime(Serializable):
    path: str
    value: datetime

    def to_proto(self, run_op: "RunOperation") -> ingest_pb2.RunOperation:
        return ingest_pb2.RunOperation(
            project=run_op.project,
            run_id=run_op.run_id,
            update=common_pb2.UpdateRunSnapshot(
                assign={
                    f"{self.path}": common_pb2.Value(
                        timestamp=timestamp_pb2.Timestamp(seconds=int(self.value.timestamp()))
                    )
                },
            ),
        )


@dataclass
class RunOperation:
    project: str
    run_id: str
    operation: Serializable

    def to_proto(self) -> ingest_pb2.RunOperation:
        return self.operation.to_proto(run_op=self)



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ("FieldToValueVisitor",)

from datetime import datetime
from typing import (
    Any,
    Optional,
    Set,
)

from neptune.api.models import (
    BoolField,
    DateTimeField,
    FieldVisitor,
    FloatField,
    FloatSeriesField,
    IntField,
    NotebookRefField,
    ObjectStateField,
    StringField,
    StringSeriesField,
    StringSetField,
)


class FieldToValueVisitor(FieldVisitor[Any]):

    def visit_float(self, field: FloatField) -> float:
        return field.value

    def visit_int(self, field: IntField) -> int:
        return field.value

    def visit_bool(self, field: BoolField) -> bool:
        return field.value

    def visit_string(self, field: StringField) -> str:
        return field.value

    def visit_datetime(self, field: DateTimeField) -> datetime:
        return field.value

    def visit_float_series(self, field: FloatSeriesField) -> Optional[float]:
        return field.last

    def visit_string_series(self, field: StringSeriesField) -> Optional[str]:
        return field.last

    def visit_string_set(self, field: StringSetField) -> Set[str]:
        return field.values

    def visit_object_state(self, field: ObjectStateField) -> str:
        return field.value

    def visit_notebook_ref(self, field: NotebookRefField) -> Optional[str]:
        return field.notebook_name



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Project"]

from typing import (
    Iterable,
    Optional,
    Union,
)

from typing_extensions import Literal

from neptune.exceptions import NeptuneUnsupportedFunctionalityException
from neptune.internal.backends.nql import NQLQuery
from neptune.internal.container_type import ContainerType
from neptune.internal.utils import (
    as_list,
    verify_collection_type,
    verify_type,
    verify_value,
)
from neptune.objects.mode import Mode
from neptune.objects.utils import (
    build_raw_query,
    prepare_nql_query,
)
from neptune.objects.with_backend import WithBackend
from neptune.table import Table
from neptune.typing import (
    ProgressBarCallback,
    ProgressBarType,
)


class Project(WithBackend):
    """Starts a connection to an existing Neptune project.

    You can use the Project object to retrieve information about runs, models, and model versions
    within the project.

    You can also log (and fetch) metadata common to the whole project, such as information about datasets,
    links to documents, or key project metrics.

    You can also use the Project object as a context manager (see examples).

    Args:
        project: Name of a project in the form `workspace-name/project-name`.
            If left empty, the value of the NEPTUNE_PROJECT environment variable is used.
        api_token: User's API token.
            If left empty, the value of the NEPTUNE_API_TOKEN environment variable is used (recommended).
        mode: Connection mode in which the tracking will work.
            If left empty, the value of the NEPTUNE_MODE environment variable is used.
            If no value was set for the environment variable, "async" is used by default.
            Possible values are `async`, `sync`, `read-only`, and `debug`.
        proxies: Argument passed to HTTP calls made via the Requests library, as dictionary of strings.

    Returns:
        Project object that can be used to interact with the project as a whole,
        like logging or fetching project-level metadata.

    Examples:

        >>> import neptune

        >>> # Connect to the project "classification" in the workspace "ml-team":
        ... project = neptune.init_project(project="ml-team/classification")

        >>> # Or initialize with the constructor
        ... project = Project(project="ml-team/classification")

        >>> # Connect to a project in read-only mode:
        ... project = neptune.init_project(
        ...     project="ml-team/classification",
        ...     mode="read-only",
        ... )

        Using the Project object as context manager:

        >>> with Project(project="ml-team/classification") as project:
        ...     project["metadata"] = some_metadata

    For more, see the docs:
        Initializing a project:
            https://docs.neptune.ai/api/neptune#init_project
        Project class reference:
            https://docs.neptune.ai/api/project/
    """

    container_type = ContainerType.PROJECT

    def __init__(
        self,
        project: Optional[str] = None,
        *,
        api_token: Optional[str] = None,
        mode: Optional[Literal["async", "sync", "read-only", "debug"]] = None,
        proxies: Optional[dict] = None,
    ):
        mode = mode or Mode.READ_ONLY.value

        if mode != Mode.READ_ONLY.value:
            raise NeptuneUnsupportedFunctionalityException

        mode = Mode(mode)

        super().__init__(
            project=project,
            api_token=api_token,
            mode=mode,
            proxies=proxies,
        )

    def _fetch_entries(
        self,
        child_type: ContainerType,
        query: NQLQuery,
        columns: Optional[Iterable[str]],
        limit: Optional[int],
        sort_by: str,
        ascending: bool,
        progress_bar: Optional[ProgressBarType],
    ) -> Table:
        if columns is not None:
            # always return entries with 'sys/id' and the column chosen for sorting when filter applied
            columns = set(columns)
            columns.add("sys/id")
            columns.add(sort_by)

        leaderboard_entries = self._backend.search_leaderboard_entries(
            project_id=self._project_id,
            types=[child_type],
            query=query,
            columns=columns,
            limit=limit,
            sort_by=sort_by,
            ascending=ascending,
            progress_bar=progress_bar,
        )

        return Table(
            backend=self._backend,
            container_type=child_type,
            entries=leaderboard_entries,
        )

    def fetch_runs_table(
        self,
        *,
        query: Optional[str] = None,
        id: Optional[Union[str, Iterable[str]]] = None,
        state: Optional[Union[Literal["inactive", "active"], Iterable[Literal["inactive", "active"]]]] = None,
        owner: Optional[Union[str, Iterable[str]]] = None,
        tag: Optional[Union[str, Iterable[str]]] = None,
        columns: Optional[Iterable[str]] = None,
        trashed: Optional[bool] = False,
        limit: Optional[int] = None,
        sort_by: str = "sys/creation_time",
        ascending: bool = False,
        progress_bar: Optional[ProgressBarType] = None,
    ) -> Table:
        """Retrieve runs matching the specified criteria.

        All parameters are optional. Each of them specifies a single criterion.
        Only runs matching all of the criteria will be returned.

        Args:
            query: NQL query string. Syntax: https://docs.neptune.ai/usage/nql/
                Example: `"(accuracy: float > 0.88) AND (loss: float < 0.2)"`.
                Exclusive with the `id`, `state`, `owner`, and `tag` parameters.
            id: Neptune ID of a run, or list of several IDs.
                Example: `"SAN-1"` or `["SAN-1", "SAN-2"]`.
                Matching any element of the list is sufficient to pass the criterion.
            state: Run state, or list of states.
                Example: `"active"`.
                Possible values: `"inactive"`, `"active"`.
                Matching any element of the list is sufficient to pass the criterion.
            owner: Username of the run owner, or a list of owners.
                Example: `"josh"` or `["frederic", "josh"]`.
                The owner is the user who created the run.
                Matching any element of the list is sufficient to pass the criterion.
            tag: A tag or list of tags applied to the run.
                Example: `"lightGBM"` or `["pytorch", "cycleLR"]`.
                Only runs that have all specified tags will match this criterion.
            columns: Names of columns to include in the table, as a list of field names.
                The Neptune ID ("sys/id") is included automatically.
                If `None` (default), all the columns of the runs table are included, up to a maximum of 10 000 columns.
            trashed: Whether to retrieve trashed runs.
                If `True`, only trashed runs are retrieved.
                If `False` (default), only not-trashed runs are retrieved.
                If `None`, both trashed and not-trashed runs are retrieved.
            limit: How many entries to return at most. If `None`, all entries are returned.
            sort_by: Name of the field to sort the results by.
                The field must represent a simple type (string, float, datetime, integer, or Boolean).
            ascending: Whether to sort the entries in ascending order of the sorting column values.
            progress_bar: Set to `False` to disable the download progress bar,
                or pass a `ProgressBarCallback` class to use your own progress bar callback.

        Returns:
            `Table` object containing `Run` objects matching the specified criteria.

            Use `to_pandas()` to convert the table to a pandas DataFrame.

        Examples:
            >>> import neptune
            ... # Fetch project "jackie/sandbox"
            ... project = neptune.init_project(mode="read-only", project="jackie/sandbox")

            >>> # Fetch the metadata of all runs as a pandas DataFrame
            ... runs_table_df = project.fetch_runs_table().to_pandas()
            ... # Extract the ID of the last run
            ... last_run_id = runs_table_df["sys/id"].values[0]

            >>> # Fetch the 100 oldest runs
            ... runs_table_df = project.fetch_runs_table(
            ...     sort_by="sys/creation_time", ascending=True, limit=100
            ... ).to_pandas()

            >>> # Fetch the 100 largest runs (space they take up in Neptune)
            ... runs_table_df = project.fetch_runs_table(sort_by="sys/size", limit=100).to_pandas()

            >>> # Include only the fields "train/loss" and "params/lr" as columns:
            ... runs_table_df = project.fetch_runs_table(columns=["params/lr", "train/loss"]).to_pandas()

            >>> # Pass a custom progress bar callback
            ... runs_table_df = project.fetch_runs_table(progress_bar=MyProgressBar).to_pandas()
            ... # The class MyProgressBar(ProgressBarCallback) must be defined

            You can also filter the runs table by state, owner, tag, or a combination of these:

            >>> # Fetch only inactive runs
            ... runs_table_df = project.fetch_runs_table(state="inactive").to_pandas()

            >>> # Fetch only runs created by CI service
            ... runs_table_df = project.fetch_runs_table(owner="my_company_ci_service").to_pandas()

            >>> # Fetch only runs that have both "Exploration" and "Optuna" tags
            ... runs_table_df = project.fetch_runs_table(tag=["Exploration", "Optuna"]).to_pandas()

            >>> # You can combine conditions. Runs satisfying all conditions will be fetched
            ... runs_table_df = project.fetch_runs_table(state="inactive", tag="Exploration").to_pandas()

        See also the API reference in the docs:
            https://docs.neptune.ai/api/project#fetch_runs_table
        """

        if any((id, state, owner, tag)) and query is not None:
            raise ValueError(
                "You can't use the 'query' parameter together with the 'id', 'state', 'owner', or 'tag' parameters."
            )

        ids = as_list("id", id)
        states = as_list("state", state)
        owners = as_list("owner", owner)
        tags = as_list("tag", tag)

        verify_type("query", query, (str, type(None)))
        verify_type("trashed", trashed, (bool, type(None)))
        verify_type("limit", limit, (int, type(None)))
        verify_type("sort_by", sort_by, str)
        verify_type("ascending", ascending, bool)
        verify_type("progress_bar", progress_bar, (type(None), bool, type(ProgressBarCallback)))
        verify_collection_type("state", states, str)

        if isinstance(limit, int) and limit <= 0:
            raise ValueError(f"Parameter 'limit' must be a positive integer or None. Got {limit}.")

        for state in states:
            verify_value("state", state.lower(), ("inactive", "active"))

        if query is not None:
            nql_query = build_raw_query(query, trashed=trashed)
        else:
            nql_query = prepare_nql_query(ids, states, owners, tags, trashed)

        return self._fetch_entries(
            child_type=ContainerType.RUN,
            query=nql_query,
            columns=columns,
            limit=limit,
            sort_by=sort_by,
            ascending=ascending,
            progress_bar=progress_bar,
        )

    def fetch_models_table(
        self,
        *,
        query: Optional[str] = None,
        columns: Optional[Iterable[str]] = None,
        trashed: Optional[bool] = False,
        limit: Optional[int] = None,
        sort_by: str = "sys/creation_time",
        ascending: bool = False,
        progress_bar: Optional[ProgressBarType] = None,
    ) -> Table:
        """Retrieve models stored in the project.

        Args:
            query: NQL query string. Syntax: https://docs.neptune.ai/usage/nql/
                Example: `"(model_size: float > 100) AND (backbone: string = VGG)"`.
            trashed: Whether to retrieve trashed models.
                If `True`, only trashed models are retrieved.
                If `False`, only not-trashed models are retrieved.
                If `None`, both trashed and not-trashed models are retrieved.
            columns: Names of columns to include in the table, as a list of field names.
                The Neptune ID ("sys/id") is included automatically.
                If `None`, all the columns of the models table are included, up to a maximum of 10 000 columns.
            limit: How many entries to return at most. If `None`, all entries are returned.
            sort_by: Name of the field to sort the results by.
                The field must represent a simple type (string, float, datetime, integer, or Boolean).
            ascending: Whether to sort the entries in ascending order of the sorting column values.
            progress_bar: Set to `False` to disable the download progress bar,
                or pass a `ProgressBarCallback` class to use your own progress bar callback.

        Returns:
            `Table` object containing `Model` objects.

            Use `to_pandas()` to convert the table to a pandas DataFrame.

        Examples:
            >>> import neptune
            ... # Fetch project "jackie/sandbox"
            ... project = neptune.init_project(mode="read-only", project="jackie/sandbox")

            >>> # Fetch the metadata of all models as a pandas DataFrame
            ... models_table_df = project.fetch_models_table().to_pandas()

            >>> # Include only the fields "params/lr" and "info/size" as columns:
            ... models_table_df = project.fetch_models_table(columns=["params/lr", "info/size"]).to_pandas()

            >>> # Fetch 10 oldest model objects
            ... models_table_df = project.fetch_models_table(
            ...     sort_by="sys/creation_time", ascending=True, limit=10
            ...  ).to_pandas()
            ... # Extract the ID of the first listed (oldest) model object
            ... last_model_id = models_table_df["sys/id"].values[0]

            >>> # Fetch models with VGG backbone
            ... models_table_df = project.fetch_models_table(
                    query="(backbone: string = VGG)"
                ).to_pandas()

        See also the API reference in the docs:
            https://docs.neptune.ai/api/project#fetch_models_table
        """
        verify_type("query", query, (str, type(None)))
        verify_type("limit", limit, (int, type(None)))
        verify_type("sort_by", sort_by, str)
        verify_type("ascending", ascending, bool)
        verify_type("progress_bar", progress_bar, (type(None), bool, type(ProgressBarCallback)))

        if isinstance(limit, int) and limit <= 0:
            raise ValueError(f"Parameter 'limit' must be a positive integer or None. Got {limit}.")

        query = query if query is not None else ""
        nql = build_raw_query(query=query, trashed=trashed)
        return self._fetch_entries(
            child_type=ContainerType.MODEL,
            query=nql,
            columns=columns,
            limit=limit,
            sort_by=sort_by,
            ascending=ascending,
            progress_bar=progress_bar,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: model/series_values.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19model/series_values.proto\x12\x11neptune.api.model\"q\n\x19ProtoFloatSeriesValuesDTO\x12\x18\n\x10total_item_count\x18\x01 \x01(\x03\x12:\n\x06values\x18\x02 \x03(\x0b\x32*.neptune.api.model.ProtoFloatPointValueDTO\"P\n\x17ProtoFloatPointValueDTO\x12\x18\n\x10timestamp_millis\x18\x01 \x01(\x03\x12\x0c\n\x04step\x18\x02 \x01(\x01\x12\r\n\x05value\x18\x03 \x01(\x01\x42\x34\n0ml.neptune.leaderboard.api.model.proto.generatedP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'model.series_values_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n0ml.neptune.leaderboard.api.model.proto.generatedP\001'
  _globals['_PROTOFLOATSERIESVALUESDTO']._serialized_start=48
  _globals['_PROTOFLOATSERIESVALUESDTO']._serialized_end=161
  _globals['_PROTOFLOATPOINTVALUEDTO']._serialized_start=163
  _globals['_PROTOFLOATPOINTVALUEDTO']._serialized_end=243
# @@protoc_insertion_point(module_scope)



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import defaultdict
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Tuple,
)

from neptune.cli.containers import (
    AsyncContainer,
    ExecutionDirectory,
    OfflineContainer,
)
from neptune.cli.utils import (
    detect_async_dir,
    detect_offline_dir,
    get_metadata_container,
    is_single_execution_dir_synced,
)
from neptune.constants import (
    ASYNC_DIRECTORY,
    OFFLINE_DIRECTORY,
)
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import UniqueId
from neptune.objects.structure_version import StructureVersion

if TYPE_CHECKING:
    from neptune.internal.backends.neptune_backend import NeptuneBackend


class CollectedContainers(NamedTuple):
    async_containers: List[AsyncContainer]
    offline_containers: List[OfflineContainer]
    synced_containers: List[AsyncContainer]
    unsynced_containers: List[AsyncContainer]
    not_found_containers: List[AsyncContainer]


def collect_containers(*, path: Path, backend: "NeptuneBackend") -> CollectedContainers:
    if not path.is_dir():
        return CollectedContainers(
            async_containers=[],
            offline_containers=[],
            synced_containers=[],
            unsynced_containers=[],
            not_found_containers=[],
        )

    async_containers: List[AsyncContainer] = []
    if (path / ASYNC_DIRECTORY).exists():
        async_containers = list(collect_async_containers(path=path, backend=backend))

    offline_containers = []
    if (path / OFFLINE_DIRECTORY).exists():
        offline_containers = list(collect_offline_containers(path=path))

    return CollectedContainers(
        async_containers=async_containers,
        offline_containers=offline_containers,
        synced_containers=[x for x in async_containers if x.synced],
        unsynced_containers=[x for x in async_containers if not x.synced and x.found is True],
        not_found_containers=[x for x in async_containers if x.found is False],
    )


def collect_async_containers(*, path: Path, backend: "NeptuneBackend") -> Iterable[AsyncContainer]:
    container_to_execution_dirs = collect_by_container(base_path=path / ASYNC_DIRECTORY, detect_by=detect_async_dir)

    for (container_type, container_id), execution_dirs in container_to_execution_dirs.items():
        experiment = get_metadata_container(backend=backend, container_type=container_type, container_id=container_id)
        found = experiment is not None

        yield AsyncContainer(
            container_id=container_id,
            container_type=container_type,
            found=found,
            experiment=experiment,
            execution_dirs=execution_dirs,
        )


def collect_offline_containers(*, path: Path) -> Iterable[OfflineContainer]:
    container_to_execution_dirs = collect_by_container(base_path=path / OFFLINE_DIRECTORY, detect_by=detect_offline_dir)

    for (container_type, container_id), execution_dirs in container_to_execution_dirs.items():
        yield OfflineContainer(
            container_id=container_id,
            container_type=container_type,
            execution_dirs=execution_dirs,
            found=False,
        )


def collect_child_directories(base_path: Path, structure_version: StructureVersion) -> List[Path]:
    if structure_version in {StructureVersion.CHILD_EXECUTION_DIRECTORIES, StructureVersion.LEGACY}:
        return list(map(lambda r: base_path / r.name, base_path.iterdir()))
    elif structure_version == StructureVersion.DIRECT_DIRECTORY:
        return [base_path]
    else:
        raise ValueError(f"Unknown structure version {structure_version}")


def collect_by_container(
    *, base_path: Path, detect_by: Callable[[str], Tuple[ContainerType, UniqueId, StructureVersion]]
) -> Dict[Tuple[ContainerType, UniqueId], List["ExecutionDirectory"]]:
    container_to_execution_dirs: Dict[Tuple[ContainerType, UniqueId], List["ExecutionDirectory"]] = defaultdict(list)

    for child_path in base_path.iterdir():
        container_type, unique_id, structure_version = detect_by(child_path.name)
        execution_dirs = collect_child_directories(child_path, structure_version)
        for execution_dir in execution_dirs:
            parent = execution_dir.parent if structure_version == StructureVersion.CHILD_EXECUTION_DIRECTORIES else None
            container_to_execution_dirs[(container_type, unique_id)].append(
                ExecutionDirectory(
                    path=execution_dir,
                    synced=is_single_execution_dir_synced(execution_dir),
                    structure_version=structure_version,
                    parent=parent,
                )
            )

    return container_to_execution_dirs



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: model/attributes.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16model/attributes.proto\x12\x11neptune.api.model\"a\n\x1eProtoAttributesSearchResultDTO\x12?\n\x07\x65ntries\x18\x01 \x03(\x0b\x32..neptune.api.model.ProtoAttributeDefinitionDTO\"9\n\x1bProtoAttributeDefinitionDTO\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\tB4\n0ml.neptune.leaderboard.api.model.proto.generatedP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'model.attributes_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n0ml.neptune.leaderboard.api.model.proto.generatedP\001'
  _globals['_PROTOATTRIBUTESSEARCHRESULTDTO']._serialized_start=45
  _globals['_PROTOATTRIBUTESSEARCHRESULTDTO']._serialized_end=142
  _globals['_PROTOATTRIBUTEDEFINITIONDTO']._serialized_start=144
  _globals['_PROTOATTRIBUTEDEFINITIONDTO']._serialized_end=201
# @@protoc_insertion_point(module_scope)



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["get_container_dir"]

import os
import random
import string
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from neptune.internal.container_type import ContainerType
    from neptune.internal.id_formats import UniqueId


RANDOM_KEY_LENGTH = 8


def get_container_dir(container_id: "UniqueId", container_type: "ContainerType") -> str:
    return f"{container_type.value}__{container_id}__{os.getpid()}__{random_key(RANDOM_KEY_LENGTH)}"


def random_key(length: int) -> str:
    characters = string.ascii_lowercase + string.digits
    return "".join(random.choice(characters) for _ in range(length))



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["OperationVisitor"]

import abc
from typing import (
    Generic,
    TypeVar,
)

from neptune.core.operations.operation import (
    AssignBool,
    AssignDatetime,
    AssignFloat,
    AssignInt,
    AssignString,
    LogFloats,
    Operation,
    RunCreation,
)

Ret = TypeVar("Ret")


class OperationVisitor(Generic[Ret]):
    def visit(self, op: Operation) -> Ret:
        return op.accept(self)

    @abc.abstractmethod
    def visit_assign_float(self, op: AssignFloat) -> Ret:
        pass

    @abc.abstractmethod
    def visit_assign_int(self, op: AssignInt) -> Ret:
        pass

    @abc.abstractmethod
    def visit_assign_bool(self, op: AssignBool) -> Ret:
        pass

    @abc.abstractmethod
    def visit_assign_string(self, op: AssignString) -> Ret:
        pass

    @abc.abstractmethod
    def visit_assign_datetime(self, op: AssignDatetime) -> Ret:
        pass

    @abc.abstractmethod
    def visit_log_floats(self, op: LogFloats) -> Ret:
        pass

    @abc.abstractmethod
    def visit_run_creation(self, op: RunCreation) -> Ret:
        pass



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["parse_iso_date"]

import datetime
from typing import Union

DATE_FORMAT_LONG: str = "%Y-%m-%dT%H:%M:%S.%fZ"
DATE_FORMAT_SHORT: str = "%Y-%m-%dT%H:%M:%SZ"


def parse_iso_date(date: Union[str, datetime.datetime]) -> datetime.datetime:
    if isinstance(date, datetime.datetime):
        return date

    date_format = DATE_FORMAT_LONG if _is_long_date_format(date) else DATE_FORMAT_SHORT

    return datetime.datetime.strptime(date, date_format)


def _is_long_date_format(date_string: str) -> bool:
    return len(date_string.split(".")) == 2



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

__all__ = ["HostedNeptuneBackendV2"]


from typing import cast

from neptune_api.api.backend import get_project
from neptune_api.credentials import Credentials
from neptune_api.models import (
    Error,
    ProjectDTO,
)
from neptune_api.types import Response

from neptune.internal.backends.api_client import (
    create_auth_api_client,
    get_config_and_token_urls,
)
from neptune.internal.id_formats import QualifiedName


class HostedNeptuneBackendV2:
    def __init__(self, credentials: Credentials) -> None:
        self.credentials = credentials

        config, token_urls = get_config_and_token_urls(self.credentials)
        self.auth_client = create_auth_api_client(credentials, config, token_urls)

    # only happy path is implemented
    def get_project(self, project_identifier: QualifiedName) -> ProjectDTO:
        response: Response[Error | ProjectDTO] = get_project.sync_detailed(
            client=self.auth_client, project_identifier=project_identifier
        )
        if response.parsed is None:
            raise RuntimeError(response.content.decode("utf-8"))
        return cast(ProjectDTO, response.parsed)



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

__all__ = (
    "Field",
    "FieldType",
    "LeaderboardEntry",
    "LeaderboardEntriesSearchResult",
    "FieldVisitor",
    "FloatField",
    "IntField",
    "BoolField",
    "StringField",
    "DateTimeField",
    "FloatSeriesField",
    "StringSeriesField",
    "StringSetField",
    "ObjectStateField",
    "NotebookRefField",
    "FieldDefinition",
    "FloatSeriesValues",
    "FloatPointValue",
    "StringSeriesValues",
    "StringPointValue",
    "QueryFieldDefinitionsResult",
    "NextPage",
    "QueryFieldsResult",
)

import abc
import re
from dataclasses import dataclass
from dataclasses import field as dataclass_field
from datetime import (
    datetime,
    timezone,
)
from enum import Enum
from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
)

from neptune.api.proto.neptune_pb.api.model.attributes_pb2 import ProtoAttributeDefinitionDTO
from neptune.api.proto.neptune_pb.api.model.leaderboard_entries_pb2 import (
    ProtoAttributesDTO,
    ProtoBoolAttributeDTO,
    ProtoDatetimeAttributeDTO,
    ProtoFloatAttributeDTO,
    ProtoFloatSeriesAttributeDTO,
    ProtoIntAttributeDTO,
    ProtoLeaderboardEntriesSearchResultDTO,
    ProtoStringAttributeDTO,
    ProtoStringSetAttributeDTO,
)
from neptune.internal.utils.iso_dates import parse_iso_date
from neptune.internal.utils.run_state import RunState

Ret = TypeVar("Ret")


class FieldType(Enum):
    FLOAT = "float"
    INT = "int"
    BOOL = "bool"
    STRING = "string"
    DATETIME = "datetime"
    FLOAT_SERIES = "floatSeries"
    STRING_SERIES = "stringSeries"
    STRING_SET = "stringSet"
    OBJECT_STATE = "experimentState"
    NOTEBOOK_REF = "notebookRef"


@dataclass
class Field(abc.ABC):
    path: str
    type: ClassVar[FieldType] = dataclass_field(init=False)
    _registry: ClassVar[Dict[str, Type[Field]]] = {}

    def __init_subclass__(cls, *args: Any, field_type: FieldType, **kwargs: Any) -> None:
        super().__init_subclass__(*args, **kwargs)
        cls.type = field_type
        cls._registry[field_type.value] = cls

    @classmethod
    def by_type(cls, field_type: FieldType) -> Type[Field]:
        return cls._registry[field_type.value]

    @abc.abstractmethod
    def accept(self, visitor: FieldVisitor[Ret]) -> Ret: ...

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> Field:
        field_type = data["type"]
        return Field._registry[field_type].from_dict(data[f"{field_type}Properties"])

    @staticmethod
    def from_model(model: Any) -> Field:
        field_type = str(model.type)
        return Field._registry[field_type].from_model(model.__getattr__(f"{field_type}Properties"))

    @staticmethod
    def from_proto(data: Any) -> Field:
        field_type = str(data.type)
        return Field._registry[field_type].from_proto(data.__getattribute__(f"{camel_to_snake(field_type)}_properties"))


def camel_to_snake(name: str) -> str:
    # Insert an underscore before any uppercase letters and convert the string to lowercase
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    # Handle the case where there are uppercase letters in the middle of the name
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


class FieldVisitor(Generic[Ret], abc.ABC):

    def visit(self, field: Field) -> Ret:
        return field.accept(self)

    @abc.abstractmethod
    def visit_float(self, field: FloatField) -> Ret: ...

    @abc.abstractmethod
    def visit_int(self, field: IntField) -> Ret: ...

    @abc.abstractmethod
    def visit_bool(self, field: BoolField) -> Ret: ...

    @abc.abstractmethod
    def visit_string(self, field: StringField) -> Ret: ...

    @abc.abstractmethod
    def visit_datetime(self, field: DateTimeField) -> Ret: ...

    @abc.abstractmethod
    def visit_float_series(self, field: FloatSeriesField) -> Ret: ...

    @abc.abstractmethod
    def visit_string_series(self, field: StringSeriesField) -> Ret: ...

    @abc.abstractmethod
    def visit_string_set(self, field: StringSetField) -> Ret: ...

    @abc.abstractmethod
    def visit_object_state(self, field: ObjectStateField) -> Ret: ...

    @abc.abstractmethod
    def visit_notebook_ref(self, field: NotebookRefField) -> Ret: ...


@dataclass
class FloatField(Field, field_type=FieldType.FLOAT):
    value: float

    def accept(self, visitor: FieldVisitor[Ret]) -> Ret:
        return visitor.visit_float(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> FloatField:
        return FloatField(path=data["attributeName"], value=float(data["value"]))

    @staticmethod
    def from_model(model: Any) -> FloatField:
        return FloatField(path=model.attributeName, value=model.value)

    @staticmethod
    def from_proto(data: ProtoFloatAttributeDTO) -> FloatField:
        return FloatField(path=data.attribute_name, value=data.value)


@dataclass
class IntField(Field, field_type=FieldType.INT):
    value: int

    def accept(self, visitor: FieldVisitor[Ret]) -> Ret:
        return visitor.visit_int(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> IntField:
        return IntField(path=data["attributeName"], value=int(data["value"]))

    @staticmethod
    def from_model(model: Any) -> IntField:
        return IntField(path=model.attributeName, value=model.value)

    @staticmethod
    def from_proto(data: ProtoIntAttributeDTO) -> IntField:
        return IntField(path=data.attribute_name, value=data.value)


@dataclass
class BoolField(Field, field_type=FieldType.BOOL):
    value: bool

    def accept(self, visitor: FieldVisitor[Ret]) -> Ret:
        return visitor.visit_bool(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> BoolField:
        return BoolField(path=data["attributeName"], value=bool(data["value"]))

    @staticmethod
    def from_model(model: Any) -> BoolField:
        return BoolField(path=model.attributeName, value=model.value)

    @staticmethod
    def from_proto(data: ProtoBoolAttributeDTO) -> BoolField:
        return BoolField(path=data.attribute_name, value=data.value)


@dataclass
class StringField(Field, field_type=FieldType.STRING):
    value: str

    def accept(self, visitor: FieldVisitor[Ret]) -> Ret:
        return visitor.visit_string(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> StringField:
        return StringField(path=data["attributeName"], value=str(data["value"]))

    @staticmethod
    def from_model(model: Any) -> StringField:
        return StringField(path=model.attributeName, value=model.value)

    @staticmethod
    def from_proto(data: ProtoStringAttributeDTO) -> StringField:
        return StringField(path=data.attribute_name, value=data.value)


@dataclass
class DateTimeField(Field, field_type=FieldType.DATETIME):
    value: datetime

    def accept(self, visitor: FieldVisitor[Ret]) -> Ret:
        return visitor.visit_datetime(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> DateTimeField:
        return DateTimeField(path=data["attributeName"], value=parse_iso_date(data["value"]))

    @staticmethod
    def from_model(model: Any) -> DateTimeField:
        return DateTimeField(path=model.attributeName, value=parse_iso_date(model.value))

    @staticmethod
    def from_proto(data: ProtoDatetimeAttributeDTO) -> DateTimeField:
        return DateTimeField(
            path=data.attribute_name, value=datetime.fromtimestamp(data.value / 1000.0, tz=timezone.utc)
        )


@dataclass
class FloatSeriesField(Field, field_type=FieldType.FLOAT_SERIES):
    last: Optional[float]

    def accept(self, visitor: FieldVisitor[Ret]) -> Ret:
        return visitor.visit_float_series(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> FloatSeriesField:
        last = float(data["last"]) if "last" in data else None
        return FloatSeriesField(path=data["attributeName"], last=last)

    @staticmethod
    def from_model(model: Any) -> FloatSeriesField:
        return FloatSeriesField(path=model.attributeName, last=model.last)

    @staticmethod
    def from_proto(data: ProtoFloatSeriesAttributeDTO) -> FloatSeriesField:
        last = data.last if data.HasField("last") else None
        return FloatSeriesField(path=data.attribute_name, last=last)


@dataclass
class StringSeriesField(Field, field_type=FieldType.STRING_SERIES):
    last: Optional[str]

    def accept(self, visitor: FieldVisitor[Ret]) -> Ret:
        return visitor.visit_string_series(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> StringSeriesField:
        last = str(data["last"]) if "last" in data else None
        return StringSeriesField(path=data["attributeName"], last=last)

    @staticmethod
    def from_model(model: Any) -> StringSeriesField:
        return StringSeriesField(path=model.attributeName, last=model.last)

    @staticmethod
    def from_proto(data: Any) -> StringSeriesField:
        raise NotImplementedError()


@dataclass
class StringSetField(Field, field_type=FieldType.STRING_SET):
    values: Set[str]

    def accept(self, visitor: FieldVisitor[Ret]) -> Ret:
        return visitor.visit_string_set(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> StringSetField:
        return StringSetField(path=data["attributeName"], values=set(map(str, data["values"])))

    @staticmethod
    def from_model(model: Any) -> StringSetField:
        return StringSetField(path=model.attributeName, values=set(model.values))

    @staticmethod
    def from_proto(data: ProtoStringSetAttributeDTO) -> StringSetField:
        return StringSetField(path=data.attribute_name, values=set(data.value))


@dataclass
class ObjectStateField(Field, field_type=FieldType.OBJECT_STATE):
    value: str

    def accept(self, visitor: FieldVisitor[Ret]) -> Ret:
        return visitor.visit_object_state(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> ObjectStateField:
        value = RunState.from_api(str(data["value"])).value
        return ObjectStateField(path=data["attributeName"], value=value)

    @staticmethod
    def from_model(model: Any) -> ObjectStateField:
        value = RunState.from_api(str(model.value)).value
        return ObjectStateField(path=model.attributeName, value=value)

    @staticmethod
    def from_proto(data: Any) -> ObjectStateField:
        raise NotImplementedError()


@dataclass
class NotebookRefField(Field, field_type=FieldType.NOTEBOOK_REF):
    notebook_name: Optional[str]

    def accept(self, visitor: FieldVisitor[Ret]) -> Ret:
        return visitor.visit_notebook_ref(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> NotebookRefField:
        notebook_name = str(data["notebookName"]) if "notebookName" in data else None
        return NotebookRefField(path=data["attributeName"], notebook_name=notebook_name)

    @staticmethod
    def from_model(model: Any) -> NotebookRefField:
        return NotebookRefField(path=model.attributeName, notebook_name=model.notebookName)

    @staticmethod
    def from_proto(data: Any) -> NotebookRefField:
        raise NotImplementedError()


@dataclass
class LeaderboardEntry:
    object_id: str
    fields: List[Field]

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> LeaderboardEntry:
        return LeaderboardEntry(
            object_id=data["experimentId"], fields=[Field.from_dict(field) for field in data["attributes"]]
        )

    @staticmethod
    def from_model(model: Any) -> LeaderboardEntry:
        return LeaderboardEntry(
            object_id=model.experimentId, fields=[Field.from_model(field) for field in model.attributes]
        )

    @staticmethod
    def from_proto(data: ProtoAttributesDTO) -> LeaderboardEntry:
        with_proto_support = {
            FieldType.STRING.value,
            FieldType.BOOL.value,
            FieldType.INT.value,
            FieldType.FLOAT.value,
            FieldType.DATETIME.value,
            FieldType.STRING_SET.value,
            FieldType.FLOAT_SERIES.value,
        }

        return LeaderboardEntry(
            object_id=data.experiment_id,
            fields=[Field.from_proto(field) for field in data.attributes if str(field.type) in with_proto_support],
        )


@dataclass
class LeaderboardEntriesSearchResult:
    entries: List[LeaderboardEntry]
    matching_item_count: int

    @staticmethod
    def from_dict(result: Dict[str, Any]) -> LeaderboardEntriesSearchResult:
        return LeaderboardEntriesSearchResult(
            entries=[LeaderboardEntry.from_dict(entry) for entry in result.get("entries", [])],
            matching_item_count=result["matchingItemCount"],
        )

    @staticmethod
    def from_model(result: Any) -> LeaderboardEntriesSearchResult:
        return LeaderboardEntriesSearchResult(
            entries=[LeaderboardEntry.from_model(entry) for entry in result.entries],
            matching_item_count=result.matchingItemCount,
        )

    @staticmethod
    def from_proto(data: ProtoLeaderboardEntriesSearchResultDTO) -> LeaderboardEntriesSearchResult:
        return LeaderboardEntriesSearchResult(
            entries=[LeaderboardEntry.from_proto(entry) for entry in data.entries],
            matching_item_count=data.matching_item_count,
        )


@dataclass
class NextPage:
    limit: Optional[int]
    next_page_token: Optional[str]

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> NextPage:
        return NextPage(limit=data.get("limit"), next_page_token=data.get("nextPageToken"))

    @staticmethod
    def from_model(model: Any) -> NextPage:
        return NextPage(limit=model.limit, next_page_token=model.nextPageToken)

    @staticmethod
    def from_proto(data: Any) -> NextPage:
        raise NotImplementedError()

    def to_dto(self) -> Dict[str, Any]:
        return {
            "limit": self.limit,
            "nextPageToken": self.next_page_token,
        }


@dataclass
class QueryFieldsExperimentResult:
    object_id: str
    object_key: str
    fields: List[Field]

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> QueryFieldsExperimentResult:
        return QueryFieldsExperimentResult(
            object_id=data["experimentId"],
            object_key=data["experimentShortId"],
            fields=[Field.from_dict(field) for field in data["attributes"]],
        )

    @staticmethod
    def from_model(model: Any) -> QueryFieldsExperimentResult:
        return QueryFieldsExperimentResult(
            object_id=model.experimentId,
            object_key=model.experimentShortId,
            fields=[Field.from_model(field) for field in model.attributes],
        )

    @staticmethod
    def from_proto(data: Any) -> QueryFieldsExperimentResult:
        raise NotImplementedError()


@dataclass
class QueryFieldsResult:
    entries: List[QueryFieldsExperimentResult]
    next_page: NextPage

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> QueryFieldsResult:
        return QueryFieldsResult(
            entries=[QueryFieldsExperimentResult.from_dict(entry) for entry in data["entries"]],
            next_page=NextPage.from_dict(data["nextPage"]),
        )

    @staticmethod
    def from_model(model: Any) -> QueryFieldsResult:
        return QueryFieldsResult(
            entries=[QueryFieldsExperimentResult.from_model(entry) for entry in model.entries],
            next_page=NextPage.from_model(model.nextPage),
        )

    @staticmethod
    def from_proto(data: Any) -> QueryFieldsResult:
        raise NotImplementedError()


@dataclass
class QueryFieldDefinitionsResult:
    entries: List[FieldDefinition]
    next_page: NextPage

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> QueryFieldDefinitionsResult:
        return QueryFieldDefinitionsResult(
            entries=[FieldDefinition.from_dict(entry) for entry in data["entries"]],
            next_page=NextPage.from_dict(data["nextPage"]),
        )

    @staticmethod
    def from_model(model: Any) -> QueryFieldDefinitionsResult:
        return QueryFieldDefinitionsResult(
            entries=[FieldDefinition.from_model(entry) for entry in model.entries],
            next_page=NextPage.from_model(model.nextPage),
        )

    @staticmethod
    def from_proto(data: Any) -> QueryFieldDefinitionsResult:
        raise NotImplementedError()


@dataclass
class FieldDefinition:
    path: str
    type: FieldType

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> FieldDefinition:
        return FieldDefinition(path=data["name"], type=FieldType(data["type"]))

    @staticmethod
    def from_model(model: Any) -> FieldDefinition:
        return FieldDefinition(path=model.name, type=FieldType(model.type))

    @staticmethod
    def from_proto(data: ProtoAttributeDefinitionDTO) -> FieldDefinition:
        return FieldDefinition(path=data.name, type=FieldType(data.type))


@dataclass
class FloatSeriesValues:
    total: int
    values: List[FloatPointValue]

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> FloatSeriesValues:
        return FloatSeriesValues(
            total=data["totalItemCount"], values=[FloatPointValue.from_dict(value) for value in data["values"]]
        )

    @staticmethod
    def from_model(model: Any) -> FloatSeriesValues:
        return FloatSeriesValues(
            total=model.totalItemCount, values=[FloatPointValue.from_model(value) for value in model.values]
        )

    @staticmethod
    def from_proto(data: Any) -> FloatSeriesValues:
        return FloatSeriesValues(
            total=data.total_item_count, values=[FloatPointValue.from_proto(value) for value in data.values]
        )


@dataclass
class FloatPointValue:
    timestamp: datetime
    value: float
    step: float

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> FloatPointValue:
        return FloatPointValue(
            timestamp=datetime.fromtimestamp(data["timestampMillis"] / 1000.0, tz=timezone.utc),
            value=float(data["value"]),
            step=float(data["step"]),
        )

    @staticmethod
    def from_model(model: Any) -> FloatPointValue:
        return FloatPointValue(
            timestamp=datetime.fromtimestamp(model.timestampMillis / 1000.0, tz=timezone.utc),
            value=model.value,
            step=model.step,
        )

    @staticmethod
    def from_proto(data: Any) -> FloatPointValue:
        return FloatPointValue(
            timestamp=datetime.fromtimestamp(data.timestamp_millis / 1000.0, tz=timezone.utc),
            value=data.value,
            step=data.step,
        )


@dataclass
class StringSeriesValues:
    total: int
    values: List[StringPointValue]

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> StringSeriesValues:
        return StringSeriesValues(
            total=data["totalItemCount"], values=[StringPointValue.from_dict(value) for value in data["values"]]
        )

    @staticmethod
    def from_model(model: Any) -> StringSeriesValues:
        return StringSeriesValues(
            total=model.totalItemCount, values=[StringPointValue.from_model(value) for value in model.values]
        )

    @staticmethod
    def from_proto(data: Any) -> StringSeriesValues:
        raise NotImplementedError()


@dataclass
class StringPointValue:
    timestamp: datetime
    step: float
    value: str

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> StringPointValue:
        return StringPointValue(
            timestamp=datetime.fromtimestamp(data["timestampMillis"] / 1000.0, tz=timezone.utc),
            value=str(data["value"]),
            step=float(data["step"]),
        )

    @staticmethod
    def from_model(model: Any) -> StringPointValue:
        return StringPointValue(
            timestamp=datetime.fromtimestamp(model.timestampMillis / 1000.0, tz=timezone.utc),
            value=model.value,
            step=model.step,
        )

    @staticmethod
    def from_proto(data: Any) -> StringPointValue:
        raise NotImplementedError()



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

__all__ = ["ProcessingResources"]

import threading
from pathlib import Path
from queue import Queue
from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Tuple,
)

from neptune.constants import ASYNC_DIRECTORY
from neptune.core.components.abstract import (
    Resource,
    WithResources,
)
from neptune.core.components.metadata_file import MetadataFile
from neptune.core.components.operation_storage import OperationStorage
from neptune.core.components.queue.aggregating_disk_queue import AggregatingDiskQueue
from neptune.core.operation_processors.utils import (
    common_metadata,
    get_container_full_path,
)
from neptune.core.operations.operation import Operation
from neptune.core.typing.container_type import ContainerType
from neptune.core.typing.id_formats import CustomId
from neptune.internal.signals_processing.signals import Signal


class ProcessingResources(WithResources):
    def __init__(
        self,
        custom_id: CustomId,
        container_type: ContainerType,
        lock: threading.RLock,
        signal_queue: Queue[Signal],
        batch_size: int = 1,
        data_path: Optional[Path] = None,
        serializer: Callable[[Operation], Dict[str, Any]] = lambda op: op.to_dict(),
    ) -> None:
        self.batch_size: int = batch_size
        self._data_path = (
            data_path if data_path else get_container_full_path(ASYNC_DIRECTORY, custom_id, container_type)
        )

        self.metadata_file = MetadataFile(
            data_path=self._data_path,
            metadata=common_metadata(mode="async", custom_id=custom_id, container_type=container_type),
        )
        self.operation_storage = OperationStorage(data_path=self._data_path)
        self.disk_queue = AggregatingDiskQueue[Operation, float](
            data_path=self._data_path,
            to_dict=serializer,
            from_dict=Operation.from_dict,
            lock=lock,
        )

        self.waiting_cond = threading.Condition()

        self.signals_queue = signal_queue

        self.consumed_version: int = 0

    @property
    def resources(self) -> Tuple[Resource, ...]:
        return self.metadata_file, self.operation_storage, self.disk_queue

    @property
    def data_path(self) -> Path:
        return self._data_path

    def cleanup(self) -> None:
        super().cleanup()
        try:
            self.data_path.rmdir()
        except OSError:
            pass



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ("SyncOperationProcessor",)

from pathlib import Path
from typing import (
    Optional,
    Tuple,
)

from neptune.constants import SYNC_DIRECTORY
from neptune.core.components.abstract import (
    Resource,
    WithResources,
)
from neptune.core.components.metadata_file import MetadataFile
from neptune.core.components.operation_storage import OperationStorage
from neptune.core.operation_processors.operation_processor import OperationProcessor
from neptune.core.operation_processors.utils import (
    common_metadata,
    get_container_full_path,
)
from neptune.core.operations.operation import Operation
from neptune.core.typing.container_type import ContainerType
from neptune.core.typing.id_formats import CustomId
from neptune.internal.utils.disk_utilization import ensure_disk_not_overutilize


class SyncOperationProcessor(WithResources, OperationProcessor):
    def __init__(self, custom_id: "CustomId", container_type: "ContainerType"):
        self._container_id: "CustomId" = custom_id
        self._container_type: "ContainerType" = container_type

        self._data_path = get_container_full_path(SYNC_DIRECTORY, custom_id, container_type)

        # Initialize directory
        self._data_path.mkdir(parents=True, exist_ok=True)

        self._metadata_file = MetadataFile(
            data_path=self._data_path,
            metadata=common_metadata(mode="sync", custom_id=custom_id, container_type=container_type),
        )
        self._operation_storage = OperationStorage(data_path=self._data_path)

    @property
    def operation_storage(self) -> "OperationStorage":
        return self._operation_storage

    @property
    def data_path(self) -> Path:
        return self._data_path

    @property
    def resources(self) -> Tuple["Resource", ...]:
        return self._metadata_file, self._operation_storage

    @ensure_disk_not_overutilize
    def enqueue_operation(self, op: "Operation", *, wait: bool) -> None: ...

    def stop(self, seconds: Optional[float] = None) -> None:
        self.flush()
        self.close()
        self.cleanup()

    def cleanup(self) -> None:
        super().cleanup()
        try:
            self._data_path.rmdir()
        except OSError:
            pass



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

__all__ = ["QueueObserver"]

from dataclasses import dataclass
from queue import Queue
from time import monotonic
from typing import (
    Optional,
    Union,
)

from neptune.core.components.queue.aggregating_disk_queue import AggregatingDiskQueue
from neptune.core.components.queue.disk_queue import DiskQueue
from neptune.core.operation_processors.async_operation_processor.consumer_thread import ConsumerThread
from neptune.core.operation_processors.async_operation_processor.operation_logger import (
    ProcessorStopLogger,
    ProcessorStopSignal,
)
from neptune.exceptions import NeptuneSynchronizationAlreadyStoppedException
from neptune.internal.utils.logger import get_logger

logger = get_logger()


@dataclass
class QueueWaitCycleResults:
    size_remaining: int
    already_synced: int
    already_synced_proc: float


class QueueObserver:
    def __init__(
        self,
        disk_queue: Union[DiskQueue, AggregatingDiskQueue],
        consumer: ConsumerThread,
        should_print_logs: bool,
        stop_queue_max_time_no_connection_seconds: float,
    ):
        self._disk_queue = disk_queue
        self._consumer = consumer
        self._stop_queue_max_time_no_connection_seconds = stop_queue_max_time_no_connection_seconds

        self._processor_stop_logger = ProcessorStopLogger(
            processor_id=id(self),
            signal_queue=None,
            logger=logger,
            should_print_logs=should_print_logs,
        )

    def is_queue_empty(self) -> bool:
        return self._disk_queue.is_empty()

    def wait_for_queue_empty(
        self,
        seconds: Optional[float],
        processor_stop_signal_queue: Optional[Queue[ProcessorStopSignal]] = None,
    ) -> None:
        initial_queue_size = self._disk_queue.size()
        waiting_start: float = monotonic()
        time_elapsed: float = 0.0
        max_reconnect_wait_time: float = self._stop_queue_max_time_no_connection_seconds if seconds is None else seconds

        self._processor_stop_logger.set_processor_stop_signal_queue(processor_stop_signal_queue)

        if initial_queue_size > 0:
            if self._consumer.last_backoff_time > 0:
                self._processor_stop_logger.log_connection_interruption(max_reconnect_wait_time)
            else:
                self._processor_stop_logger.log_remaining_operations(size_remaining=initial_queue_size)

        while True:
            wait_cycle_results = self._wait_single_cycle(
                seconds,
                self._processor_stop_logger,
                initial_queue_size,
                waiting_start,
                time_elapsed,
                max_reconnect_wait_time,
            )
            if wait_cycle_results is None:
                # either there are no more operations to process or there is a synchronization failure
                return
            self._processor_stop_logger.log_still_waiting(
                size_remaining=wait_cycle_results.size_remaining,
                already_synced=wait_cycle_results.already_synced,
                already_synced_proc=wait_cycle_results.already_synced_proc,
            )

    def _wait_single_cycle(
        self,
        seconds: Optional[float],
        op_logger: ProcessorStopLogger,
        initial_queue_size: int,
        waiting_start: float,
        time_elapsed: float,
        max_reconnect_wait_time: float,
    ) -> Optional[QueueWaitCycleResults]:
        if seconds is None:
            if self._consumer.last_backoff_time == 0:
                # reset `waiting_start` on successful action
                waiting_start = monotonic()
            wait_time = self._stop_queue_max_time_no_connection_seconds
        else:
            wait_time = max(
                min(
                    seconds - time_elapsed,
                    self._stop_queue_max_time_no_connection_seconds,
                ),
                0.0,
            )
        self._disk_queue.wait_for_empty(wait_time)

        cycle_results = _calculate_wait_cycle_results(self._disk_queue, initial_queue_size)

        if cycle_results.size_remaining == 0:
            op_logger.log_success(ops_synced=initial_queue_size)
            return None

        time_elapsed = monotonic() - waiting_start
        if self._consumer.last_backoff_time > 0 and time_elapsed >= max_reconnect_wait_time:
            op_logger.log_reconnect_failure(
                max_reconnect_wait_time=max_reconnect_wait_time,
                size_remaining=cycle_results.size_remaining,
            )
            return None

        if seconds is not None and wait_time == 0:
            op_logger.log_sync_failure(seconds=seconds, size_remaining=cycle_results.size_remaining)
            return None

        if not self._consumer.is_running():
            exception = NeptuneSynchronizationAlreadyStoppedException()
            logger.warning(str(exception))
            return None

        return cycle_results


def _calculate_wait_cycle_results(
    disk_queue: Union[DiskQueue, AggregatingDiskQueue], initial_queue_size: int
) -> QueueWaitCycleResults:
    size_remaining = disk_queue.size()
    already_synced = initial_queue_size - size_remaining
    already_synced_proc = (already_synced / initial_queue_size) * 100 if initial_queue_size else 100

    return QueueWaitCycleResults(size_remaining, already_synced, already_synced_proc)



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["TokenRefreshingURLs", "get_config_and_token_urls", "create_auth_api_client"]


from dataclasses import dataclass
from typing import Tuple

from neptune_api import (
    AuthenticatedClient,
    Client,
)
from neptune_api.api.backend import get_client_config
from neptune_api.auth_helpers import exchange_api_key
from neptune_api.credentials import Credentials
from neptune_api.models import (
    ClientConfig,
    Error,
)


@dataclass
class TokenRefreshingURLs:
    authorization_endpoint: str
    token_endpoint: str

    @classmethod
    def from_dict(cls, data: dict) -> "TokenRefreshingURLs":
        return TokenRefreshingURLs(
            authorization_endpoint=data["authorization_endpoint"], token_endpoint=data["token_endpoint"]
        )


def get_config_and_token_urls(credentials: Credentials) -> Tuple[ClientConfig, TokenRefreshingURLs]:
    with Client(base_url=credentials.base_url) as client:
        config = get_client_config.sync(client=client)
        if config is None or isinstance(config, Error):
            raise RuntimeError(f"Failed to get client config: {config}")
        response = client.get_httpx_client().get(config.security.open_id_discovery)
        token_urls = TokenRefreshingURLs.from_dict(response.json())
    return config, token_urls


def create_auth_api_client(
    credentials: Credentials, config: ClientConfig, token_refreshing_urls: TokenRefreshingURLs
) -> AuthenticatedClient:
    return AuthenticatedClient(
        base_url=credentials.base_url,
        credentials=credentials,
        client_id=config.security.client_id,
        token_refreshing_endpoint=token_refreshing_urls.token_endpoint,
        api_key_exchange_callback=exchange_api_key,
    )
